export type Event_T = [name:string, payload:any];
export interface Machine_T {
  states: Array<State_T>
}
export interface State_T {
  name: string;
  ons: Array<On_T>;
}
export interface On_T {
  eventName: string;
  reactions: Array<Do_T | Goto_T>;
};
export interface Do_T {
  type: 'Do';
  fn: DoFn_T;
};
export type DoFn_T = (ctx:any,e:Event_T,self:Interpreter_T)=>void;
export interface Goto_T {
  type: 'Goto';
  targetStateName: string;
};

export const Machine = function(...states:Array<State_T>) : Machine_T { return {states}; };
export const State = function(name:string, ...ons:Array<On_T>) : State_T{ return {name, ons}; };
export const On = function(eventName:string, ...reactions:Array<Do_T | Goto_T>) : On_T{ return {eventName, reactions}; };
export const Do = function<S,E extends Event_T, C>(fn:DoFn_T) : Do_T{ return {type:'Do', fn}; };
export const Goto = function(targetStateName:string) : Goto_T { return {type:'Goto', targetStateName} };

interface Tick_T {
  doFunctions: Array<DoFn_T>
};
const Tick = function(doFunctions : Array<DoFn_T>) : Tick_T{ return {doFunctions}; };


export interface Interpreter_T {
  machine: Machine_T;
  state: string;
  context: any;
  tickQueues:Array<TickQueue_T>
}
type TickQueue_T = Array<Tick_T>;
export function interpret(machine:Machine_T, options:{state?:string, context:any}) : Interpreter_T{
  let {state, context} = options;
  if(typeof state === 'undefined'){ state = machine.states[0].name; }
  const interpreter = {machine, state, context, tickQueues:[]}
  console.log(interpreter);
  //@ts-ignore
  send(interpreter, ['entry', null] );
  return interpreter;
}

/**  Helper function for `send()`
 */
function getState(interpreter : Interpreter_T) : State_T{
  return interpreter.machine.states.find((state)=>state.name===interpreter.state) as unknown as State_T;
}
/**  Helper function for `send()`
 */
function getOns(state : State_T, event:Event_T) : Array<On_T>{
  return state.ons.filter((on)=>on.eventName===event[0]);
}
/**  Helper function for `send()`
 */
function noop(){}
/**  Inject an Event into the Interpreter's "tick queue".
 * 
 * An event can be signify something "new" happening, such that its reactions should run on the next Tick;
 * or it can signify a milestone "within" the current Tick, such that a Tick can be thought of as having 
 * "sub-Ticks".
 * 
 * This distinction is significant for proper ordering of reaction execution, and also for determining
 * whether to run a reaction at all. If an Event is received, and is specified to be applied on a past 
 * Tick, it is discarded.
 */
export function send(interpreter : Interpreter_T, event:Event_T){
  const state = getState(interpreter);
  const onsTree = getOns(state, event);
  const reactions = onsTree
    .map((on)=>on.reactions)
    .flat();
  const indexOfFirstGoto = reactions.findIndex((reaction)=>reaction.type==='Goto');
  const indexOfFinalReaction = indexOfFirstGoto === -1 ? reactions.length-1 : indexOfFirstGoto;
  const reactionsUntilFirstGoto = reactions.slice(0, indexOfFinalReaction+1);
  const functionsToRunInTick = reactionsUntilFirstGoto
    .map((reaction)=>{
      if(reaction.type === 'Do'){
        return reaction.fn;
      }
      else if(reaction.type === 'Goto'){
        return (ctx:any,e:Event_T,self:Interpreter_T)=>{
          self.state = reaction.targetStateName;
          //@ts-ignore
          send(self, ['entry', e] );
        };
      }
      else{
        return noop;
      }
    });
  const tick = Tick(functionsToRunInTick);
  tick.doFunctions.forEach((fn)=>{ fn(interpreter.context, event, interpreter); });
}

export const Spawn = function(){};
export const Unspawn = function(){};